@@ -34,6 +34,11 @@
 #include <libgen.h>
 #include <inttypes.h>
 #include <limits.h>
+#include <sys/stat.h>
+#include <dirent.h>
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 
 #include <libimobiledevice/libimobiledevice.h>
 #include <libimobiledevice/lockdown.h>
@@ -45,30 +50,45 @@
 
 #include <zip.h>
 
+#ifndef ZIP_CODEC_ENCODE
+// this is required for old libzip...
+#define zip_get_num_entries(x, y) zip_get_num_files(x)
+#define zip_int64_t ssize_t
+#define zip_uint64_t off_t
+#endif
+
+#define ITUNES_METADATA_PLIST_FILENAME "iTunesMetadata.plist"
+
 const char PKG_PATH[] = "PublicStaging";
 const char APPARCH_PATH[] = "ApplicationArchives";
 
-char *uuid = NULL;
+char *udid = NULL;
 char *options = NULL;
 char *appid = NULL;
 
-int list_apps_mode = 0;
-int install_mode = 0;
-int uninstall_mode = 0;
-int upgrade_mode = 0;
-int list_archives_mode = 0;
-int archive_mode = 0;
-int restore_mode = 0;
-int remove_archive_mode = 0;
+enum cmd_mode {
+	CMD_NONE = 0,
+	CMD_LIST_APPS,
+	CMD_INSTALL,
+	CMD_UNINSTALL,
+	CMD_UPGRADE,
+	CMD_LIST_ARCHIVES,
+	CMD_ARCHIVE,
+	CMD_RESTORE,
+	CMD_REMOVE_ARCHIVE
+};
+
+int cmd = CMD_NONE;
 
 char *last_status = NULL;
 int wait_for_op_complete = 0;
 int notification_expected = 0;
+int is_device_connected = 0;
 int op_completed = 0;
-int err_occured = 0;
+int err_occurred = 0;
 int notified = 0;
 
-#ifdef HAVE_LIBIMOBILEDEVICE_1_0
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 static void notifier(const char *notification, void *unused)
 #else
 static void notifier(const char *notification)
@@ -103,20 +123,20 @@ static void status_cb(const char *operation, plist_t status)
 		}
 		if (!nerror) {
 			if (last_status && (strcmp(last_status, status_msg))) {
-				printf("\n");
+				printf("\r");
 			}
 
 			if (!npercent) {
 				printf("%s - %s\n", operation, status_msg);
 			} else {
-				printf("%s - %s (%d%%)\r", operation, status_msg, percent);
+				printf("%s - %s (%d%%)\n", operation, status_msg, percent);
 			}
 		} else {
 			char *err_msg = NULL;
 			plist_get_string_val(nerror, &err_msg);
-			printf("%s - Error occured: %s\n", operation, err_msg);
+			printf("%s - Error occurred: %s\n", operation, err_msg);
 			free(err_msg);
-			err_occured = 1;
+			err_occurred = 1;
 		}
 		if (last_status) {
 			free(last_status);
@@ -131,7 +151,7 @@ static void status_cb(const char *operation, plist_t status)
 	}
 }
 
-static int zip_f_get_contents(struct zip *zf, const char *filename, int locate_flags, char **buffer, uint32_t *len)
+static int zip_get_contents(struct zip *zf, const char *filename, int locate_flags, char **buffer, uint32_t *len)
 {
 	struct zip_stat zs;
 	struct zip_file *zfile;
@@ -141,7 +161,6 @@ static int zip_f_get_contents(struct zip *zf, const char *filename, int locate_f
 	*len = 0;
 
 	if (zindex < 0) {
-		fprintf(stderr, "ERROR: could not locate %s in archive!\n", filename);
 		return -1;
 	}
 
@@ -176,25 +195,109 @@ static int zip_f_get_contents(struct zip *zf, const char *filename, int locate_f
 	return 0;
 }
 
-static void do_wait_when_needed()
+static int zip_get_app_directory(struct zip* zf, char** path)
 {
 	int i = 0;
+	int c = zip_get_num_files(zf);
+	int len = 0;
+	const char* name = NULL;
+
+	/* look through all filenames in the archive */
+	do {
+		/* get filename at current index */
+		name = zip_get_name(zf, i++, 0);
+		if (name != NULL) {
+			/* check if we have a "Payload/.../" name */
+			len = strlen(name);
+			if (!strncmp(name, "Payload/", 8) && (len > 8)) {
+				/* locate the second directory delimiter */
+				const char* p = name + 8;
+				do {
+					if (*p == '/') {
+						break;
+					}
+				} while(p++ != NULL);
+
+				/* try next entry if not found */
+				if (p == NULL)
+					continue;
+
+				len = p - name + 1;
+
+				if (*path != NULL) {
+					free(*path);
+					*path = NULL;
+				}
+
+				/* allocate and copy filename */
+				*path = (char*)malloc(len + 1);
+				strncpy(*path, name, len);
+
+				/* add terminating null character */
+				char* t = *path + len;
+				*t = '\0';
+				break;
+			}
+		}
+	} while(i < c);
+
+	return 0;
+}
+
+static void idevice_event_callback(const idevice_event_t* event, void* userdata)
+{
+	if (event->event == IDEVICE_DEVICE_REMOVE) {
+		is_device_connected = 0;
+	}
+}
+
+static void idevice_wait_for_operation_to_complete()
+{
 	struct timespec ts;
 	ts.tv_sec = 0;
 	ts.tv_nsec = 500000000;
+	is_device_connected = 1;
+
+	/* subscribe to make sure to exit on device removal */
+	idevice_event_subscribe(idevice_event_callback, NULL);
 
 	/* wait for operation to complete */
-	while (wait_for_op_complete && !op_completed && !err_occured
-		   && !notified && (i < 60)) {
+	while (wait_for_op_complete && !op_completed && !err_occurred
+		   && !notified && is_device_connected) {
 		nanosleep(&ts, NULL);
-		i++;
 	}
 
 	/* wait some time if a notification is expected */
-	while (notification_expected && !notified && !err_occured && (i < 10)) {
+	while (notification_expected && !notified && !err_occurred && is_device_connected) {
 		nanosleep(&ts, NULL);
-		i++;
 	}
+
+	idevice_event_unsubscribe();
+}
+
+static int str_is_udid(const char* str)
+{
+	const char allowed[] = "0123456789abcdefABCDEF";
+
+	/* handle NULL case */
+	if (str == NULL)
+		return -1;
+
+	int length = strlen(str);
+
+	/* verify length */
+	if (length != 40)
+		return -1;
+
+	/* check for invalid characters */
+	while(length--) {
+		/* invalid character in udid? */
+		if (strchr(allowed, str[length]) == NULL) {
+			return -1;
+		}
+	} 
+
+	return 0;
 }
 
 static void print_usage(int argc, char **argv)
@@ -203,22 +306,24 @@ static void print_usage(int argc, char **argv)
 
 	name = strrchr(argv[0], '/');
 	printf("Usage: %s OPTIONS\n", (name ? name + 1 : argv[0]));
-	printf("Manage apps on an iDevice.\n\n");
+	printf("Manage apps on iOS devices.\n\n");
 	printf
-		("  -U, --uuid UUID\tTarget specific device by its 40-digit device UUID.\n"
+		("  -u, --udid UDID\tTarget specific device by its 40-digit device UDID.\n"
 		 "  -l, --list-apps\tList apps, possible options:\n"
 		 "       -o list_user\t- list user apps only (this is the default)\n"
 		 "       -o list_system\t- list system apps only\n"
 		 "       -o list_all\t- list all types of apps\n"
 		 "       -o xml\t\t- print full output as xml plist\n"
 		 "  -i, --install ARCHIVE\tInstall app from package file specified by ARCHIVE.\n"
-		 "  -u, --uninstall APPID\tUninstall app specified by APPID.\n"
-		 "  -g, --upgrade APPID\tUpgrade app specified by APPID.\n"
+		 "                       \tARCHIVE can also be a .ipcc file for carrier bundles.\n"
+		 "  -U, --uninstall APPID\tUninstall app specified by APPID.\n"
+		 "  -g, --upgrade ARCHIVE\tUpgrade app from package file specified by ARCHIVE.\n"
 		 "  -L, --list-archives\tList archived applications, possible options:\n"
 		 "       -o xml\t\t- print full output as xml plist\n"
 		 "  -a, --archive APPID\tArchive app specified by APPID, possible options:\n"
 		 "       -o uninstall\t- uninstall the package after making an archive\n"
 		 "       -o app_only\t- archive application data only\n"
+		 "       -o docs_only\t- archive documents (user data) only\n"
 		 "       -o copy=PATH\t- copy the app archive to directory PATH when done\n"
 		 "       -o remove\t- only valid when copy=PATH is used: remove after copy\n"
 		 "  -r, --restore APPID\tRestore archived app specified by APPID\n"
@@ -232,10 +337,10 @@ static void parse_opts(int argc, char **argv)
 {
 	static struct option longopts[] = {
 		{"help", 0, NULL, 'h'},
-		{"uuid", 1, NULL, 'U'},
+		{"udid", 1, NULL, 'u'},
 		{"list-apps", 0, NULL, 'l'},
 		{"install", 1, NULL, 'i'},
-		{"uninstall", 1, NULL, 'u'},
+		{"uninstall", 1, NULL, 'U'},
 		{"upgrade", 1, NULL, 'g'},
 		{"list-archives", 0, NULL, 'L'},
 		{"archive", 1, NULL, 'a'},
@@ -254,47 +359,77 @@ static void parse_opts(int argc, char **argv)
 			break;
 		}
 
+		/* verify if multiple modes have been supplied */
+		switch (c) {
+		case 'l':
+		case 'i':
+		case 'g':
+		case 'L':
+		case 'a':
+		case 'r':
+		case 'R':
+			if (cmd != CMD_NONE) {
+				printf("ERROR: A mode has already been supplied. Multiple modes are not supported.\n");
+				print_usage(argc, argv);
+				exit(2);
+			}
+			break;
+		default:
+			break;
+		}
+
 		switch (c) {
 		case 'h':
 			print_usage(argc, argv);
 			exit(0);
-		case 'U':
-			if (strlen(optarg) != 40) {
-				printf("%s: invalid UUID specified (length != 40)\n",
-					   argv[0]);
+		case 'u':
+			if (str_is_udid(optarg) == 0) {
+				udid = strdup(optarg);
+				break;
+			}
+			if (strchr(optarg, '.') != NULL) {
+				fprintf(stderr, "WARNING: Using \"-u\" for \"--uninstall\" is deprecated. Please use \"-U\" instead.\n");
+				cmd = CMD_UNINSTALL;
+				appid = strdup(optarg);
+			} else {
+				printf("ERROR: Invalid UDID specified\n");
 				print_usage(argc, argv);
 				exit(2);
 			}
-			uuid = strdup(optarg);
 			break;
 		case 'l':
-			list_apps_mode = 1;
+			cmd = CMD_LIST_APPS;
 			break;
 		case 'i':
-			install_mode = 1;
+			cmd = CMD_INSTALL;
 			appid = strdup(optarg);
 			break;
-		case 'u':
-			uninstall_mode = 1;
+		case 'U':
+			if (str_is_udid(optarg) == 0) {
+				fprintf(stderr, "WARNING: Using \"-U\" for \"--udid\" is deprecated. Please use \"-u\" instead.\n");
+				udid = strdup(optarg);
+				break;
+			}
+			cmd = CMD_UNINSTALL;
 			appid = strdup(optarg);
 			break;
 		case 'g':
-			upgrade_mode = 1;
+			cmd = CMD_UPGRADE;
 			appid = strdup(optarg);
 			break;
 		case 'L':
-			list_archives_mode = 1;
+			cmd = CMD_LIST_ARCHIVES;
 			break;
 		case 'a':
-			archive_mode = 1;
+			cmd = CMD_ARCHIVE;
 			appid = strdup(optarg);
 			break;
 		case 'r':
-			restore_mode = 1;
+			cmd = CMD_RESTORE;
 			appid = strdup(optarg);
 			break;
 		case 'R':
-			remove_archive_mode = 1;
+			cmd = CMD_REMOVE_ARCHIVE;
 			appid = strdup(optarg);
 			break;
 		case 'o':
@@ -318,29 +453,134 @@ static void parse_opts(int argc, char **argv)
 		}
 	}
 
-	if (optind <= 1 || (argc - optind > 0)) {
+	if (cmd == CMD_NONE) {
+		printf("ERROR: No mode/operation was supplied.\n");
+	}
+
+	if (cmd == CMD_NONE || optind <= 1 || (argc - optind > 0)) {
 		print_usage(argc, argv);
 		exit(2);
 	}
 }
 
+static int afc_upload_file(afc_client_t afc, const char* filename, const char* dstfn)
+{
+	FILE *f = NULL;
+	uint64_t af = 0;
+	char buf[1048576];
+
+	f = fopen(filename, "rb");
+	if (!f) {
+		fprintf(stderr, "fopen: %s: %s\n", appid, strerror(errno));
+		return -1;
+	}
+
+	if ((afc_file_open(afc, dstfn, AFC_FOPEN_WRONLY, &af) != AFC_E_SUCCESS) || !af) {
+		fclose(f);
+		fprintf(stderr, "afc_file_open on '%s' failed!\n", dstfn);
+		return -1;
+	}
+
+	size_t amount = 0;
+	do {
+		amount = fread(buf, 1, sizeof(buf), f);
+		if (amount > 0) {
+			uint32_t written, total = 0;
+			while (total < amount) {
+				written = 0;
+				afc_error_t aerr = afc_file_write(afc, af, buf, amount, &written);
+				if (aerr != AFC_E_SUCCESS) {
+					fprintf(stderr, "AFC Write error: %d\n", aerr);
+					break;
+				}
+				total += written;
+			}
+			if (total != amount) {
+				fprintf(stderr, "Error: wrote only %d of %zu\n", total, amount);
+				afc_file_close(afc, af);
+				fclose(f);
+				return -1;
+			}
+		}
+	} while (amount > 0);
+
+	afc_file_close(afc, af);
+	fclose(f);
+
+	return 0;
+}
+
+static void afc_upload_dir(afc_client_t afc, const char* path, const char* afcpath)
+{
+	afc_make_directory(afc, afcpath);
+
+	DIR *dir = opendir(path);
+	if (dir) {
+		struct dirent* ep;
+		while ((ep = readdir(dir))) {
+			if ((strcmp(ep->d_name, ".") == 0) || (strcmp(ep->d_name, "..") == 0)) {
+				continue;
+			}
+			char *fpath = (char*)malloc(strlen(path)+1+strlen(ep->d_name)+1);
+			char *apath = (char*)malloc(strlen(afcpath)+1+strlen(ep->d_name)+1);
+
+			struct stat st;
+
+			strcpy(fpath, path);
+			strcat(fpath, "/");
+			strcat(fpath, ep->d_name);
+
+			strcpy(apath, afcpath);
+			strcat(apath, "/");
+			strcat(apath, ep->d_name);
+
+#ifdef HAVE_LSTAT
+			if ((lstat(fpath, &st) == 0) && S_ISLNK(st.st_mode)) {
+				char *target = (char *)malloc(st.st_size+1);
+				if (readlink(fpath, target, st.st_size+1) < 0) {
+					fprintf(stderr, "ERROR: readlink: %s (%d)\n", strerror(errno), errno);
+				} else {
+					target[st.st_size] = '\0';
+					afc_make_link(afc, AFC_SYMLINK, target, fpath);
+				}
+				free(target);
+			} else
+#endif
+			if ((stat(fpath, &st) == 0) && S_ISDIR(st.st_mode)) {
+				afc_upload_dir(afc, fpath, apath);
+			} else {
+				afc_upload_file(afc, fpath, apath);
+			}
+			free(fpath);
+			free(apath);
+		}
+		closedir(dir);
+	}
+}
+
 int main(int argc, char **argv)
 {
 	idevice_t phone = NULL;
 	lockdownd_client_t client = NULL;
 	instproxy_client_t ipc = NULL;
+	instproxy_error_t err;
 	np_client_t np = NULL;
 	afc_client_t afc = NULL;
-	uint16_t port = 0;
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+	lockdownd_service_descriptor_t service = NULL;
+#else
+	uint16_t service = 0;
+#endif
 	int res = 0;
+	char *bundleidentifier = NULL;
 
 	parse_opts(argc, argv);
 
 	argc -= optind;
 	argv += optind;
 
-	if (IDEVICE_E_SUCCESS != idevice_new(&phone, uuid)) {
-		fprintf(stderr, "No iPhone found, is it plugged in?\n");
+	if (IDEVICE_E_SUCCESS != idevice_new(&phone, udid)) {
+		fprintf(stderr, "No iOS device found, is it plugged in?\n");
 		return -1;
 	}
 
@@ -351,18 +591,27 @@ int main(int argc, char **argv)
 
 	if ((lockdownd_start_service
 		 (client, "com.apple.mobile.notification_proxy",
-		  &port) != LOCKDOWN_E_SUCCESS) || !port) {
+		  &service) != LOCKDOWN_E_SUCCESS) || !service) {
 		fprintf(stderr,
 				"Could not start com.apple.mobile.notification_proxy!\n");
 		goto leave_cleanup;
 	}
 
-	if (np_client_new(phone, port, &np) != NP_E_SUCCESS) {
+	np_error_t nperr = np_client_new(phone, service, &np);
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+	if (service) {
+		lockdownd_service_descriptor_free(service);
+	}
+	service = NULL;
+#else
+	service = 0;
+#endif
+	if (nperr != NP_E_SUCCESS) {
 		fprintf(stderr, "Could not connect to notification_proxy!\n");
 		goto leave_cleanup;
 	}
 
-#ifdef HAVE_LIBIMOBILEDEVICE_1_0
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1
 	np_set_notify_callback(np, notifier, NULL);
 #else
 	np_set_notify_callback(np, notifier);
@@ -373,16 +622,31 @@ int main(int argc, char **argv)
 	np_observe_notifications(np, noties);
 
 run_again:
-	port = 0;
-	if ((lockdownd_start_service
-		 (client, "com.apple.mobile.installation_proxy",
-		  &port) != LOCKDOWN_E_SUCCESS) || !port) {
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+	if (service) {
+		lockdownd_service_descriptor_free(service);
+	}
+	service = NULL;
+#else
+	service = 0;
+#endif
+	if ((lockdownd_start_service(client, "com.apple.mobile.installation_proxy",
+		  &service) != LOCKDOWN_E_SUCCESS) || !service) {
 		fprintf(stderr,
 				"Could not start com.apple.mobile.installation_proxy!\n");
 		goto leave_cleanup;
 	}
 
-	if (instproxy_client_new(phone, port, &ipc) != INSTPROXY_E_SUCCESS) {
+	err = instproxy_client_new(phone, service, &ipc);
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+	if (service) {
+		lockdownd_service_descriptor_free(service);
+	}
+	service = NULL;
+#else
+	service = 0;
+#endif
+	if (err != INSTPROXY_E_SUCCESS) {
 		fprintf(stderr, "Could not connect to installation_proxy!\n");
 		goto leave_cleanup;
 	}
@@ -395,11 +659,10 @@ run_again:
 	}
 	notification_expected = 0;
 
-	if (list_apps_mode) {
+	if (cmd == CMD_LIST_APPS) {
 		int xml_mode = 0;
 		plist_t client_opts = instproxy_client_options_new();
 		instproxy_client_options_add(client_opts, "ApplicationType", "User", NULL);
-		instproxy_error_t err;
 		plist_t apps = NULL;
 
 		/* look for options */
@@ -422,6 +685,7 @@ run_again:
 				}
 				elem = strtok(NULL, ",");
 			}
+			free(opts);
 		}
 
 		err = instproxy_browse(ipc, client_opts, &apps);
@@ -488,18 +752,24 @@ run_again:
 			free(s_appid);
 		}
 		plist_free(apps);
-	} else if (install_mode || upgrade_mode) {
+	} else if (cmd == CMD_INSTALL || cmd == CMD_UPGRADE) {
 		plist_t sinf = NULL;
 		plist_t meta = NULL;
 		char *pkgname = NULL;
 		struct stat fst;
-		FILE *f = NULL;
 		uint64_t af = 0;
 		char buf[8192];
 
-		port = 0;
-		if ((lockdownd_start_service(client, "com.apple.afc", &port) !=
-			 LOCKDOWN_E_SUCCESS) || !port) {
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+		if (service) {
+			lockdownd_service_descriptor_free(service);
+		}
+		service = NULL;
+#else
+		service = 0;
+#endif
+		if ((lockdownd_start_service(client, "com.apple.afc", &service) !=
+			 LOCKDOWN_E_SUCCESS) || !service) {
 			fprintf(stderr, "Could not start com.apple.afc!\n");
 			goto leave_cleanup;
 		}
@@ -507,7 +777,7 @@ run_again:
 		lockdownd_client_free(client);
 		client = NULL;
 
-		if (afc_client_new(phone, port, &afc) != INSTPROXY_E_SUCCESS) {
+		if (afc_client_new(phone, service, &afc) != AFC_E_SUCCESS) {
 			fprintf(stderr, "Could not connect to AFC!\n");
 			goto leave_cleanup;
 		}
@@ -517,98 +787,6 @@ run_again:
 			goto leave_cleanup;
 		}
 
-		/* open install package */
-		int errp = 0;
-		struct zip *zf = zip_open(appid, 0, &errp);
-		if (!zf) {
-			fprintf(stderr, "ERROR: zip_open: %s: %d\n", appid, errp);
-			goto leave_cleanup;
-		}
-
-		/* extract iTunesMetadata.plist from package */
-		char *zbuf = NULL;
-		uint32_t len = 0;
-		if (zip_f_get_contents(zf, "iTunesMetadata.plist", 0, &zbuf, &len) == 0) {
-			meta = plist_new_data(zbuf, len);
-		}
-		if (zbuf) {
-			free(zbuf);
-		}
-
-		/* we need to get the CFBundleName first */
-		plist_t info = NULL;
-		zbuf = NULL;
-		len = 0;
-		if (zip_f_get_contents(zf, "Info.plist", ZIP_FL_NODIR, &zbuf, &len) < 0) {
-			zip_unchange_all(zf);
-			zip_close(zf);
-			goto leave_cleanup;
-		}
-		if (memcmp(zbuf, "bplist00", 8) == 0) {
-			plist_from_bin(zbuf, len, &info);
-		} else {
-			plist_from_xml(zbuf, len, &info);
-		}
-		free(zbuf);
-
-		if (!info) {
-			fprintf(stderr, "Could not parse Info.plist!\n");
-			zip_unchange_all(zf);
-			zip_close(zf);
-			goto leave_cleanup;
-		}
-
-		char *bundlename = NULL;
-
-		plist_t bname = plist_dict_get_item(info, "CFBundleName");
-		if (bname) {
-			plist_get_string_val(bname, &bundlename);
-		}
-		plist_free(info);
-
-		if (!bundlename) {
-			fprintf(stderr, "Could not determine CFBundleName!\n");
-			zip_unchange_all(zf);
-			zip_close(zf);
-			goto leave_cleanup;
-		}
-
-		char *sinfname = NULL;
-	       	if (asprintf(&sinfname, "Payload/%s.app/SC_Info/%s.sinf", bundlename, bundlename) < 0) {
-			fprintf(stderr, "Out of memory!?\n");
-			goto leave_cleanup;
-		}
-		free(bundlename);
-
-		/* extract .sinf from package */
-		zbuf = NULL;
-		len = 0;
-		if (zip_f_get_contents(zf, sinfname, 0, &zbuf, &len) == 0) {
-			sinf = plist_new_data(zbuf, len);
-		}
-		free(sinfname);
-		if (zbuf) {
-			free(zbuf);
-		}
-
-		zip_unchange_all(zf);
-		zip_close(zf);
-
-		/* copy archive to device */
-		f = fopen(appid, "r");
-		if (!f) {
-			fprintf(stderr, "fopen: %s: %s\n", appid, strerror(errno));
-			goto leave_cleanup;
-		}
-
-		pkgname = NULL;
-		if (asprintf(&pkgname, "%s/%s", PKG_PATH, basename(appid)) < 0) {
-			fprintf(stderr, "Out of memory!?\n");
-			goto leave_cleanup;
-		}
-
-		printf("Copying '%s' --> '%s'\n", appid, pkgname);
-
 		char **strs = NULL;
 		if (afc_get_file_info(afc, PKG_PATH, &strs) != AFC_E_SUCCESS) {
 			if (afc_make_directory(afc, PKG_PATH) != AFC_E_SUCCESS) {
@@ -624,62 +802,265 @@ run_again:
 			free(strs);
 		}
 
-		if ((afc_file_open(afc, pkgname, AFC_FOPEN_WRONLY, &af) !=
-			 AFC_E_SUCCESS) || !af) {
-			fclose(f);
-			fprintf(stderr, "afc_file_open on '%s' failed!\n", pkgname);
-			free(pkgname);
-			goto leave_cleanup;
-		}
+		plist_t client_opts = instproxy_client_options_new();
 
-		size_t amount = 0;
-		do {
-			amount = fread(buf, 1, sizeof(buf), f);
-			if (amount > 0) {
-				uint32_t written, total = 0;
-				while (total < amount) {
-					written = 0;
-					if (afc_file_write(afc, af, buf, amount, &written) !=
-						AFC_E_SUCCESS) {
-						fprintf(stderr, "AFC Write error!\n");
-						break;
+		/* open install package */
+		int errp = 0;
+		struct zip *zf = NULL;
+		
+		if ((strlen(appid) > 5) && (strcmp(&appid[strlen(appid)-5], ".ipcc") == 0)) {
+			zf = zip_open(appid, 0, &errp);
+			if (!zf) {
+				fprintf(stderr, "ERROR: zip_open: %s: %d\n", appid, errp);
+				goto leave_cleanup;
+			}
+
+			char* ipcc = strdup(appid);
+			if ((asprintf(&pkgname, "%s/%s", PKG_PATH, basename(ipcc)) > 0) && pkgname) {
+				afc_make_directory(afc, pkgname);
+			}
+
+			printf("Uploading %s package contents... ", basename(ipcc));
+
+			/* extract the contents of the .ipcc file to PublicStaging/<name>.ipcc directory */
+			zip_uint64_t numzf = zip_get_num_entries(zf, 0);
+			zip_uint64_t i = 0;
+			for (i = 0; numzf > 0 && i < numzf; i++) {
+				const char* zname = zip_get_name(zf, i, 0);
+				char* dstpath = NULL;
+				if (!zname) continue;
+				if (zname[strlen(zname)-1] == '/') {
+					// directory
+					if ((asprintf(&dstpath, "%s/%s/%s", PKG_PATH, basename(ipcc), zname) > 0) && dstpath) {
+						afc_make_directory(afc, dstpath);						}
+					free(dstpath);
+					dstpath = NULL;
+				} else {
+					// file
+					struct zip_file* zfile = zip_fopen_index(zf, i, 0);
+					if (!zfile) continue;
+
+					if ((asprintf(&dstpath, "%s/%s/%s", PKG_PATH, basename(ipcc), zname) <= 0) || !dstpath || (afc_file_open(afc, dstpath, AFC_FOPEN_WRONLY, &af) != AFC_E_SUCCESS)) {
+						fprintf(stderr, "ERROR: can't open afc://%s for writing\n", dstpath);
+						free(dstpath);
+						dstpath = NULL;
+						zip_fclose(zfile);
+						continue;
 					}
-					total += written;
-				}
-				if (total != amount) {
-					fprintf(stderr, "Error: wrote only %d of %zu\n", total,
-							amount);
+
+					struct zip_stat zs;
+					zip_stat_init(&zs);
+					if (zip_stat_index(zf, i, 0, &zs) != 0) {
+						fprintf(stderr, "ERROR: zip_stat_index %" PRIu64 " failed!\n", i);
+						free(dstpath);
+						dstpath = NULL;
+						zip_fclose(zfile);
+						continue;
+					}
+
+					free(dstpath);
+					dstpath = NULL;
+
+					zip_uint64_t zfsize = 0;
+					while (zfsize < zs.size) {
+						zip_int64_t amount = zip_fread(zfile, buf, sizeof(buf));
+						if (amount == 0) {
+							break;
+						}
+
+						if (amount > 0) {
+							uint32_t written, total = 0;
+							while (total < amount) {
+								written = 0;
+								if (afc_file_write(afc, af, buf, amount, &written) !=
+									AFC_E_SUCCESS) {
+									fprintf(stderr, "AFC Write error!\n");
+									break;
+								}
+								total += written;
+							}
+							if (total != amount) {
+								fprintf(stderr, "Error: wrote only %d of %" PRIi64 "\n", total, amount);
+								afc_file_close(afc, af);
+								zip_fclose(zfile);
+								free(dstpath);
+								goto leave_cleanup;
+							}
+						}
+
+						zfsize += amount;
+					}
+
 					afc_file_close(afc, af);
-					fclose(f);
-					free(pkgname);
-					goto leave_cleanup;
+					af = 0;
+
+					zip_fclose(zfile);
 				}
 			}
+			free(ipcc);
+			printf("DONE.\n");
+
+			instproxy_client_options_add(client_opts, "PackageType", "CarrierBundle", NULL);
+		} else if (S_ISDIR(fst.st_mode)) {
+			/* upload developer app directory */
+			instproxy_client_options_add(client_opts, "PackageType", "Developer", NULL);
+
+			if (asprintf(&pkgname, "%s/%s", PKG_PATH, basename(appid)) < 0) {
+				fprintf(stderr, "ERROR: Out of memory allocating pkgname!?\n");
+				goto leave_cleanup;
+			}
+
+			printf("Uploading %s package contents... ", basename(appid));
+			afc_upload_dir(afc, appid, pkgname);
+			printf("DONE.\n");
+		} else {
+			zf = zip_open(appid, 0, &errp);
+			if (!zf) {
+				fprintf(stderr, "ERROR: zip_open: %s: %d\n", appid, errp);
+				goto leave_cleanup;
+			}
+
+			/* extract iTunesMetadata.plist from package */
+			char *zbuf = NULL;
+			uint32_t len = 0;
+			plist_t meta_dict = NULL;
+			if (zip_get_contents(zf, ITUNES_METADATA_PLIST_FILENAME, 0, &zbuf, &len) == 0) {
+				meta = plist_new_data(zbuf, len);
+				if (memcmp(zbuf, "bplist00", 8) == 0) {
+					plist_from_bin(zbuf, len, &meta_dict);
+				} else {
+					plist_from_xml(zbuf, len, &meta_dict);
+				}
+			} else {
+				fprintf(stderr, "WARNING: could not locate %s in archive!\n", ITUNES_METADATA_PLIST_FILENAME);
+			}
+			if (zbuf) {
+				free(zbuf);
+			}
+
+			/* determine .app directory in archive */
+			zbuf = NULL;
+			len = 0;
+			plist_t info = NULL;
+			char* filename = NULL;
+			char* app_directory_name = NULL;
+
+			if (zip_get_app_directory(zf, &app_directory_name)) {
+				fprintf(stderr, "Unable to locate app directory in archive!\n");
+				goto leave_cleanup;
+			}
+
+			/* construct full filename to Info.plist */
+			filename = (char*)malloc(strlen(app_directory_name)+10+1);
+			strcpy(filename, app_directory_name);
+			free(app_directory_name);
+			app_directory_name = NULL;
+			strcat(filename, "Info.plist");
+
+			if (zip_get_contents(zf, filename, 0, &zbuf, &len) < 0) {
+				fprintf(stderr, "WARNING: could not locate %s in archive!\n", filename);
+				free(filename);
+				zip_unchange_all(zf);
+				zip_close(zf);
+				goto leave_cleanup;
+			}
+			free(filename);
+			if (memcmp(zbuf, "bplist00", 8) == 0) {
+				plist_from_bin(zbuf, len, &info);
+			} else {
+				plist_from_xml(zbuf, len, &info);
+			}
+			free(zbuf);
+
+			if (!info) {
+				fprintf(stderr, "Could not parse Info.plist!\n");
+				zip_unchange_all(zf);
+				zip_close(zf);
+				goto leave_cleanup;
+			}
+
+			char *bundleexecutable = NULL;
+
+			plist_t bname = plist_dict_get_item(info, "CFBundleExecutable");
+			if (bname) {
+				plist_get_string_val(bname, &bundleexecutable);
+			}
+
+			bname = plist_dict_get_item(info, "CFBundleIdentifier");
+			if (bname) {
+				plist_get_string_val(bname, &bundleidentifier);
+			}
+			plist_free(info);
+			info = NULL;
+
+			if (!bundleexecutable) {
+				fprintf(stderr, "Could not determine value for CFBundleExecutable!\n");
+				zip_unchange_all(zf);
+				zip_close(zf);
+				goto leave_cleanup;
+			}
+
+			char *sinfname = NULL;
+			if (asprintf(&sinfname, "Payload/%s.app/SC_Info/%s.sinf", bundleexecutable, bundleexecutable) < 0) {
+				fprintf(stderr, "Out of memory!?\n");
+				goto leave_cleanup;
+			}
+			free(bundleexecutable);
+
+			/* extract .sinf from package */
+			zbuf = NULL;
+			len = 0;
+			if (zip_get_contents(zf, sinfname, 0, &zbuf, &len) == 0) {
+				sinf = plist_new_data(zbuf, len);
+			} else {
+				fprintf(stderr, "WARNING: could not locate %s in archive!\n", sinfname);
+			}
+			free(sinfname);
+			if (zbuf) {
+				free(zbuf);
+			}
+
+			/* copy archive to device */
+			pkgname = NULL;
+			if (asprintf(&pkgname, "%s/%s", PKG_PATH, bundleidentifier) < 0) {
+				fprintf(stderr, "Out of memory!?\n");
+				goto leave_cleanup;
+			}
+
+			printf("Copying '%s' to device... ", appid);
+
+			if (afc_upload_file(afc, appid, pkgname) < 0) {
+				free(pkgname);
+				goto leave_cleanup;
+			}
+
+			printf("DONE.\n");
+
+			if (bundleidentifier) {
+				instproxy_client_options_add(client_opts, "CFBundleIdentifier", bundleidentifier, NULL);
+			}
+			if (sinf) {
+				instproxy_client_options_add(client_opts, "ApplicationSINF", sinf, NULL);
+			}
+			if (meta) {
+				instproxy_client_options_add(client_opts, "iTunesMetadata", meta, NULL);
+			}
+		}
+		if (zf) {
+			zip_unchange_all(zf);
+			zip_close(zf);
 		}
-		while (amount > 0);
-
-		afc_file_close(afc, af);
-		fclose(f);
-
-		printf("done.\n");
 
 		/* perform installation or upgrade */
-		plist_t client_opts = instproxy_client_options_new();
-		if (sinf) {
-			instproxy_client_options_add(client_opts, "ApplicationSINF", sinf, NULL);
-		}
-		if (meta) {
-			instproxy_client_options_add(client_opts, "iTunesMetadata", meta, NULL);
-		}
-		if (install_mode) {
-			printf("Installing '%s'\n", pkgname);
+		if (cmd == CMD_INSTALL) {
+			printf("Installing '%s'\n", bundleidentifier);
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 			instproxy_install(ipc, pkgname, client_opts, status_cb, NULL);
 #else
 			instproxy_install(ipc, pkgname, client_opts, status_cb);
 #endif
 		} else {
-			printf("Upgrading '%s'\n", pkgname);
+			printf("Upgrading '%s'\n", bundleidentifier);
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 			instproxy_upgrade(ipc, pkgname, client_opts, status_cb, NULL);
 #else
@@ -690,19 +1071,19 @@ run_again:
 		free(pkgname);
 		wait_for_op_complete = 1;
 		notification_expected = 1;
-	} else if (uninstall_mode) {
+	} else if (cmd == CMD_UNINSTALL) {
+		printf("Uninstalling '%s'\n", appid);
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 		instproxy_uninstall(ipc, appid, NULL, status_cb, NULL);
 #else
 		instproxy_uninstall(ipc, appid, NULL, status_cb);
 #endif
 		wait_for_op_complete = 1;
-		notification_expected = 1;
-	} else if (list_archives_mode) {
+		notification_expected = 0;
+	} else if (cmd == CMD_LIST_ARCHIVES) {
 		int xml_mode = 0;
 		plist_t dict = NULL;
 		plist_t lres = NULL;
-		instproxy_error_t err;
 
 		/* look for options */
 		if (options) {
@@ -789,11 +1170,12 @@ run_again:
 		}
 		while (node);
 		plist_free(dict);
-	} else if (archive_mode) {
+	} else if (cmd == CMD_ARCHIVE) {
 		char *copy_path = NULL;
 		int remove_after_copy = 0;
 		int skip_uninstall = 1;
 		int app_only = 0;
+		int docs_only = 0;
 		plist_t client_opts = NULL;
 
 		/* look for options */
@@ -805,6 +1187,10 @@ run_again:
 					skip_uninstall = 0;
 				} else if (!strcmp(elem, "app_only")) {
 					app_only = 1;
+					docs_only = 0;
+				} else if (!strcmp(elem, "docs_only")) {
+					docs_only = 1;
+					app_only = 0;
 				} else if ((strlen(elem) > 5) && !strncmp(elem, "copy=", 5)) {
 					copy_path = strdup(elem+5);
 				} else if (!strcmp(elem, "remove")) {
@@ -814,13 +1200,15 @@ run_again:
 			}
 		}
 
-		if (skip_uninstall || app_only) {
+		if (skip_uninstall || app_only || docs_only) {
 			client_opts = instproxy_client_options_new();
 			if (skip_uninstall) {
 				instproxy_client_options_add(client_opts, "SkipUninstall", 1, NULL);
 			}
 			if (app_only) {
 				instproxy_client_options_add(client_opts, "ArchiveType", "ApplicationOnly", NULL);
+			} else if (docs_only) {
+				instproxy_client_options_add(client_opts, "ArchiveType", "DocumentsOnly", NULL);
 			}
 		}
 
@@ -838,8 +1226,15 @@ run_again:
 				goto leave_cleanup;
 			}
 
-			port = 0;
-			if ((lockdownd_start_service(client, "com.apple.afc", &port) != LOCKDOWN_E_SUCCESS) || !port) {
+#ifdef HAVE_LIBIMOBILEDEVICE_1_1_5
+			if (service) {
+				lockdownd_service_descriptor_free(service);
+			}
+			service = NULL;
+#else
+			service = 0;
+#endif
+			if ((lockdownd_start_service(client, "com.apple.afc", &service) != LOCKDOWN_E_SUCCESS) || !service) {
 				fprintf(stderr, "Could not start com.apple.afc!\n");
 				free(copy_path);
 				goto leave_cleanup;
@@ -848,7 +1243,7 @@ run_again:
 			lockdownd_client_free(client);
 			client = NULL;
 
-			if (afc_client_new(phone, port, &afc) != INSTPROXY_E_SUCCESS) {
+			if (afc_client_new(phone, service, &afc) != AFC_E_SUCCESS) {
 				fprintf(stderr, "Could not connect to AFC!\n");
 				goto leave_cleanup;
 			}
@@ -867,10 +1262,10 @@ run_again:
 			notification_expected = 1;
 		}
 
-		do_wait_when_needed();
+		idevice_wait_for_operation_to_complete();
 
 		if (copy_path) {
-			if (err_occured) {
+			if (err_occurred) {
 				afc_client_free(afc);
 				afc = NULL;
 				goto leave_cleanup;
@@ -885,7 +1280,7 @@ run_again:
 			}
 			free(copy_path);
 
-			f = fopen(localfile, "w");
+			f = fopen(localfile, "wb");
 			if (!f) {
 				fprintf(stderr, "ERROR: fopen: %s: %s\n", localfile, strerror(errno));
 				free(localfile);
@@ -940,7 +1335,7 @@ run_again:
 			}
 
 			/* copy file over */
-			printf("Copying '%s' --> '%s'\n", remotefile, localfile);
+			printf("Copying '%s' --> '%s'... ", remotefile, localfile);
 			free(remotefile);
 			free(localfile);
 
@@ -967,7 +1362,8 @@ run_again:
 			afc_file_close(afc, af);
 			fclose(f);
 
-			printf("done.\n");
+			printf("DONE.\n");
+
 			if (total != fsize) {
 				fprintf(stderr, "WARNING: remote and local file sizes don't match (%d != %d)\n", fsize, total);
 				if (remove_after_copy) {
@@ -979,8 +1375,7 @@ run_again:
 			if (remove_after_copy) {
 				/* remove archive if requested */
 				printf("Removing '%s'\n", appid);
-				archive_mode = 0;
-				remove_archive_mode = 1;
+				cmd = CMD_REMOVE_ARCHIVE;
 				free(options);
 				options = NULL;
 				if (LOCKDOWN_E_SUCCESS != lockdownd_client_new_with_handshake(phone, &client, "ideviceinstaller")) {
@@ -991,7 +1386,7 @@ run_again:
 			}
 		}
 		goto leave_cleanup;
-	} else if (restore_mode) {
+	} else if (cmd == CMD_RESTORE) {
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 		instproxy_restore(ipc, appid, NULL, status_cb, NULL);
 #else
@@ -999,7 +1394,7 @@ run_again:
 #endif
 		wait_for_op_complete = 1;
 		notification_expected = 1;
-	} else if (remove_archive_mode) {
+	} else if (cmd == CMD_REMOVE_ARCHIVE) {
 #ifdef HAVE_LIBIMOBILEDEVICE_1_1
 		instproxy_remove_archive(ipc, appid, NULL, status_cb, NULL);
 #else
@@ -1019,9 +1414,12 @@ run_again:
 		client = NULL;
 	}
 
-	do_wait_when_needed();
+	idevice_wait_for_operation_to_complete();
 
-  leave_cleanup:
+leave_cleanup:
+	if (bundleidentifier) {
+		free(bundleidentifier);
+	}
 	if (np) {
 		np_client_free(np);
 	}
@@ -1036,8 +1434,8 @@ run_again:
 	}
 	idevice_free(phone);
 
-	if (uuid) {
-		free(uuid);
+	if (udid) {
+		free(udid);
 	}
 	if (appid) {
 		free(appid);
@@ -1046,5 +1444,9 @@ run_again:
 		free(options);
 	}
 
+	if (err_occurred && !res) {
+		res = 128;
+	}
+
 	return res;
 }
